"""Common classes for Plugboard schemas."""

from abc import ABC

from pydantic import BaseModel, ConfigDict


class PlugboardBaseModel(BaseModel, ABC):
    """Custom base model for Plugboard schemas."""

    model_config = ConfigDict(
        extra="forbid", populate_by_name=True, use_enum_values=True, validate_assignment=True
    )



"""Provides definitions for entity names and id syntax."""

from enum import StrEnum
import typing as _t


_ENTITY_SEP: _t.Final[str] = "_"
_ENTITY_ID_MIN_CHARS: _t.Final[int] = 8
_ENTITY_ID_SUFFIX_REGEX: _t.Final[str] = (
    _ENTITY_SEP + rf"(?P<id>[a-zA-Z0-9]{{{_ENTITY_ID_MIN_CHARS},}})$"
)


class Entity(StrEnum):
    """Entity names."""

    Job = "Job"

    @property
    def id_prefix(self) -> str:
        """Returns prefix for generating unique entity ids."""
        return str(self) + _ENTITY_SEP

    @property
    def id_regex(self) -> str:
        """Returns regex for validating entity ids."""
        return rf"^(?P<entity>{self})" + _ENTITY_ID_SUFFIX_REGEX


ENTITY_ID_REGEX = rf"^(?P<entity>{'|'.join([e.name for e in Entity])})" + _ENTITY_ID_SUFFIX_REGEX



"""Provides schemas used in Plugboard.

This includes:

* Pydantic models for specifying Plugboard objects;
* `TypeDict` definitions for constructor `**kwargs`.
"""

from importlib.metadata import version

from ._common import PlugboardBaseModel
from .component import ComponentArgsDict, ComponentArgsSpec, ComponentSpec
from .config import ConfigSpec, ProcessConfigSpec
from .connector import (
    DEFAULT_CONNECTOR_CLS_PATH,
    ConnectorBuilderArgsDict,
    ConnectorBuilderArgsSpec,
    ConnectorBuilderSpec,
    ConnectorMode,
    ConnectorSocket,
    ConnectorSpec,
)
from .entities import ENTITY_ID_REGEX, Entity
from .io import IODirection
from .process import ProcessArgsDict, ProcessArgsSpec, ProcessSpec
from .state import StateBackendArgsDict, StateBackendArgsSpec, StateBackendSpec, Status
from .tune import (
    CategoricalParameterSpec,
    Direction,
    FloatParameterSpec,
    IntParameterSpec,
    ObjectiveSpec,
    OptunaSpec,
    ParameterSpec,
    TuneArgsDict,
    TuneArgsSpec,
    TuneSpec,
)


__version__ = version(__package__)


__all__ = [
    "CategoricalParameterSpec",
    "ComponentSpec",
    "ComponentArgsDict",
    "ComponentArgsSpec",
    "ConfigSpec",
    "ConnectorBuilderArgsDict",
    "ConnectorBuilderArgsSpec",
    "ConnectorBuilderSpec",
    "ConnectorMode",
    "ConnectorSocket",
    "ConnectorSpec",
    "DEFAULT_CONNECTOR_CLS_PATH",
    "Direction",
    "ENTITY_ID_REGEX",
    "Entity",
    "FloatParameterSpec",
    "IntParameterSpec",
    "IODirection",
    "ObjectiveSpec",
    "OptunaSpec",
    "ParameterSpec",
    "PlugboardBaseModel",
    "ProcessConfigSpec",
    "ProcessSpec",
    "ProcessArgsDict",
    "ProcessArgsSpec",
    "StateBackendSpec",
    "StateBackendArgsDict",
    "StateBackendArgsSpec",
    "Status",
    "TuneArgsDict",
    "TuneArgsSpec",
    "TuneSpec",
]



"""Provides top-level `ConfigSpec` class for Plugboard configuration."""

from pathlib import Path
import typing as _t

import msgspec
from pydantic import field_validator

from ._common import PlugboardBaseModel
from .process import ProcessSpec
from .tune import TuneSpec


class ProcessConfigSpec(PlugboardBaseModel):
    """A `ProcessSpec` within a Plugboard configuration.

    Attributes:
        process: A `ProcessSpec` that specifies the process, or a path to a YAML file containing the
            process specification.
        tune: Optional; A `TuneSpec` that specifies an optimisation configuration.
    """

    process: ProcessSpec
    tune: TuneSpec | None = None

    @field_validator("process", mode="before")
    @classmethod
    def _auto_load_process(cls, value: _t.Any) -> _t.Any:
        """Automatically loads the process specification from a YAML file if a path is provided."""
        if isinstance(value, str) and Path(value).exists():
            with open(value, "rb") as f:
                other_config = msgspec.yaml.decode(f.read())
            try:
                return other_config["plugboard"]["process"]
            except KeyError:
                raise ValueError(
                    "The provided YAML file does not contain a Plugboard process specification."
                )
        return value


class ConfigSpec(PlugboardBaseModel):
    """Configuration for a Plugboard simulation.

    Attributes:
        plugboard: A `ProcessConfig` that specifies the Plugboard `Process`.
    """

    plugboard: ProcessConfigSpec



"""Provides `ProcessSpec` class."""

import typing as _t

from annotated_types import Len
from pydantic import field_validator, model_validator
from typing_extensions import Self

from ._common import PlugboardBaseModel
from .component import ComponentSpec
from .connector import DEFAULT_CONNECTOR_CLS_PATH, ConnectorBuilderSpec, ConnectorSpec
from .state import StateBackendSpec


class ProcessArgsDict(_t.TypedDict):
    """`TypedDict` of the [`Process`][plugboard.process.Process] constructor arguments."""

    components: list[ComponentSpec]
    connectors: list[ConnectorSpec]
    name: _t.NotRequired[str | None]
    parameters: dict[str, _t.Any]
    state: _t.NotRequired[StateBackendSpec | None]


class ProcessArgsSpec(PlugboardBaseModel, extra="allow"):
    """Specification of the [`Process`][plugboard.process.Process] constructor arguments.

    Attributes:
        components: Specifies each `Component` in the `Process`.
        connectors: Specifies the connections between each `Component`.
        name: Unique identifier for `Process`.
        parameters: Parameters for the `Process`. These will be shared across all `Component`
            objects within the `Process`.
        state: Optional; Specifies the `StateBackend` used for the `Process`.
    """

    components: _t.Annotated[list[ComponentSpec], Len(min_length=1)]
    connectors: list[ConnectorSpec] = []
    name: _t.Optional[str] = None
    parameters: dict[str, _t.Any] = {}
    state: StateBackendSpec = StateBackendSpec()


class ProcessSpec(PlugboardBaseModel):
    """Specification of a Plugboard [`Process`][plugboard.process.Process].

    Attributes:
        args: The arguments for the `Process`.
        type: The type of `Process` to build.
        connector_builder: The `ConnectorBuilder` to use for the `Process`.
    """

    args: ProcessArgsSpec
    type: _t.Literal[
        "plugboard.process.LocalProcess",
        "plugboard.process.RayProcess",
    ] = "plugboard.process.LocalProcess"
    connector_builder: ConnectorBuilderSpec = ConnectorBuilderSpec()

    @model_validator(mode="after")
    def _validate_channel_builder_type(self: Self) -> Self:
        if (
            self.type.endswith("RayProcess")
            and self.connector_builder.type == DEFAULT_CONNECTOR_CLS_PATH
        ):
            raise ValueError("RayProcess requires a parallel-capable connector type.")
        return self

    @field_validator("type", mode="before")
    @classmethod
    def _validate_type(cls, value: _t.Any) -> str:
        if isinstance(value, str):
            return {
                "plugboard.process.local_process.LocalProcess": "plugboard.process.LocalProcess",
                "plugboard.process.ray_process.RayProcess": "plugboard.process.RayProcess",
            }.get(value, value)
        return value



"""Provides classes related to IO."""

from enum import StrEnum


class IODirection(StrEnum):
    """`IODirection` defines the type of IO operation.

    Attributes:
        INPUT: Specifies an input to a [`Component`][plugboard.component.Component].
        OUTPUT: Specifies an output to a [`Component`][plugboard.component.Component].
    """

    INPUT = "input"
    OUTPUT = "output"



"""Provides `StateBackendSpec` class."""

from datetime import datetime, timezone
from enum import StrEnum
import typing as _t

from pydantic import Field

from ._common import PlugboardBaseModel
from .entities import Entity


DEFAULT_STATE_BACKEND_CLS_PATH: str = "plugboard.state.DictStateBackend"


class Status(StrEnum):
    """`Status` describes the status of either a `Component` or a `Process`.

    Attributes:
        CREATED: The `Component` or `Process` has been created but not yet started.
        INIT: The `Component` or `Process` has been initialised but has not started running.
        RUNNING: The `Component` or `Process` is currently running.
        WAITING: The `Component` or `Process` is waiting for input.
        COMPLETED: The `Component` or `Process` has completed successfully.
        FAILED: The `Component` or `Process` has failed.
        STOPPED: The `Component` or `Process` has been cancelled or stopped.
    """

    CREATED = "created"
    INIT = "init"
    RUNNING = "running"
    WAITING = "waiting"
    COMPLETED = "completed"
    FAILED = "failed"
    STOPPED = "stopped"

    @property
    def is_terminal(self) -> bool:
        """Returns whether the status is terminal."""
        return self in {self.COMPLETED, self.FAILED, self.STOPPED}


class StateBackendArgsDict(_t.TypedDict):
    """`TypedDict` of the [`StateBackend`][plugboard.state.StateBackend] constructor arguments."""

    job_id: _t.NotRequired[str | None]
    metadata: _t.NotRequired[dict[str, _t.Any] | None]


class StateBackendArgsSpec(PlugboardBaseModel, extra="allow"):
    """Specification of the [`StateBackend`][plugboard.state.StateBackend] constructor arguments.

    Attributes:
        job_id: The unique id for the job.
        metadata: Metadata for a run.
    """

    job_id: _t.Optional[str] = Field(default=None, pattern=Entity.Job.id_regex)
    metadata: dict[str, _t.Any] = {}


class StateBackendSpec(PlugboardBaseModel):
    """Specification of a Plugboard [`StateBackend`][plugboard.state.StateBackend].

    Attributes:
        type: The type of the `StateBackend`.
        args: The arguments for the `StateBackend`.
    """

    type: str = DEFAULT_STATE_BACKEND_CLS_PATH
    args: StateBackendArgsSpec = StateBackendArgsSpec()


class StateSchema(PlugboardBaseModel):
    """Schema for Plugboard state data."""

    job_id: str = Field(pattern=Entity.Job.id_regex)
    created_at: str = Field(default_factory=lambda: datetime.now(timezone.utc).isoformat())
    metadata: dict = {}



"""Provides `ComponentSpec` class."""

import typing as _t

from pydantic import Field

from ._common import PlugboardBaseModel


class ComponentArgsDict(_t.TypedDict):
    """`TypedDict` of the [`Component`][plugboard.component.Component] constructor arguments."""

    name: str
    initial_values: _t.NotRequired[dict[str, _t.Any] | None]
    parameters: _t.NotRequired[dict[str, _t.Any] | None]
    constraints: _t.NotRequired[dict[str, _t.Any] | None]


class ComponentArgsSpec(PlugboardBaseModel, extra="allow"):
    """Specification of the [`Component`][plugboard.component.Component] constructor arguments.

    Attributes:
        name: The name of the `Component`.
        initial_values: Initial values for the `Component`.
        parameters: Parameters for the `Component`.
        constraints: Constraints for the `Component`.
    """

    name: str = Field(pattern=r"^([a-zA-Z_][a-zA-Z0-9_-]*)$")
    initial_values: dict[str, _t.Any] = {}
    parameters: dict[str, _t.Any] = {}
    constraints: dict[str, _t.Any] = {}


class ComponentSpec(PlugboardBaseModel):
    """Specification of a [`Component`][plugboard.component.Component].

    Attributes:
        type: The type of the `Component`.
        args: The arguments for the `Component`.
    """

    type: str
    args: ComponentArgsSpec



"""Provides the `TuneSpec` class for configuring optimisation jobs."""

from abc import ABC
import typing as _t

from pydantic import Field, PositiveInt, ValidationInfo, field_validator, model_validator

from ._common import PlugboardBaseModel


class OptunaSpec(PlugboardBaseModel):
    """Specification for the Optuna configuration.

    See: https://docs.ray.io/en/latest/tune/api/doc/ray.tune.search.optuna.OptunaSearch.html
    and https://optuna.readthedocs.io/en/stable/reference/index.html for more information on the
    Optuna configuration.

    Attributes:
        type: The algorithm type to load.
        space: Optional; A function defining the search space. Use this to define more complex
            search spaces that cannot be represented using the built-in parameter types.
        study_name: Optional; The name of the study.
        storage: Optional; The storage URI to save the optimisation results to.
    """

    type: _t.Literal["ray.tune.search.optuna.OptunaSearch"] = "ray.tune.search.optuna.OptunaSearch"
    space: str | None = None
    study_name: str | None = None
    storage: str | None = None


class BaseFieldSpec(PlugboardBaseModel, ABC):
    """Base class for specifying fields within a Plugboard [`Process`][plugboard.process.Process].

    These fields may be used as adjustable parameter inputs or as an optimisation objective.

    Attributes:
        object_type: The type of object on which the field is defined. Defaults to "component".
        object_name: Optional; The name of the object on which the field is defined.
        field_type: The type of field. This can be "arg", "initial_value", or "field".
        field_name: The name of the field.
    """

    object_type: _t.Literal["component", "process"] = Field("component", exclude=True)
    object_name: str | None = Field(default=None, exclude=True)
    field_type: _t.Literal["arg", "initial_value", "field", "parameter"] = Field(..., exclude=True)
    field_name: str = Field(..., exclude=True)

    @field_validator("object_name")
    @classmethod
    def _validate_object_name(cls, v: str | None, info: ValidationInfo) -> str | None:
        if info.data.get("object_type") == "component" and v is None:  # pragma: no cover
            raise ValueError("Component name must be specified.")
        return v

    @field_validator("field_type")
    @classmethod
    def _validate_field_type(cls, v: str, info: ValidationInfo) -> str:
        if info.data.get("object_type") == "process":  # pragma: no cover
            if v != "parameter":
                raise ValueError("If object_type is 'process', field_type must be 'parameter'.")
        return v

    @property
    def full_name(self) -> str:
        """Returns the full name of the field, including the object name and field name."""
        return (
            f"{self.object_type}."
            f"{self.object_name if self.object_name else 'default'}."
            f"{self.field_type}.{self.field_name}"
        )


class ObjectiveSpec(BaseFieldSpec):
    """Specification for an objective field."""

    @model_validator(mode="before")
    @classmethod
    def _fill_defaults(
        cls, data: dict[str, _t.Any] | list[dict[str, _t.Any]]
    ) -> dict[str, _t.Any] | list[dict[str, _t.Any]]:
        if isinstance(data, list):
            # If the data is a list, skip because it is already a list of objectives
            return data
        if "field_type" not in data:
            data["field_type"] = "field"
        if data["field_type"] != "field":  # pragma: no cover
            raise ValueError("The field type must be 'field' for an objective specification.")
        return data


class FloatParameterSpec(BaseFieldSpec):
    """Specification for a uniform float parameter.

    See: https://docs.ray.io/en/latest/tune/api/search_space.html.

    Attributes:
        type: The type of the parameter.
        lower: The lower bound of the parameter.
        upper: The upper bound of the parameter.
    """

    type: _t.Literal["ray.tune.uniform"] = "ray.tune.uniform"
    lower: float
    upper: float


class IntParameterSpec(BaseFieldSpec):
    """Specification for a uniform integer parameter.

    See: https://docs.ray.io/en/latest/tune/api/search_space.html.

    Attributes:
        type: The type of the parameter.
        lower: The lower bound of the parameter.
        upper: The upper bound of the parameter.
    """

    type: _t.Literal["ray.tune.randint"] = "ray.tune.randint"
    lower: int
    upper: int


class CategoricalParameterSpec(BaseFieldSpec):
    """Specification for a categorical parameter.

    See: https://docs.ray.io/en/latest/tune/api/search_space.html.

    Attributes:
        type: The type of the parameter.
        categories: The categories of the parameter.
    """

    type: _t.Literal["ray.tune.choice"] = "ray.tune.choice"
    categories: list[_t.Any]


ParameterSpec = _t.Union[
    FloatParameterSpec,
    IntParameterSpec,
    CategoricalParameterSpec,
]
"""A union type for all parameter specifications."""

Direction = _t.Literal["min", "max"]
"""A type for the direction of optimisation."""


class TuneArgsDict(_t.TypedDict):
    """`TypedDict` of the [`Tuner`][plugboard.tune.Tuner] constructor arguments."""

    objective: str | list[str]
    parameters: list[ParameterSpec]
    num_samples: int
    mode: _t.NotRequired[Direction | list[Direction]]
    max_concurrent: _t.NotRequired[int | None]
    algorithm: OptunaSpec


class TuneArgsSpec(PlugboardBaseModel):
    """Specification of the arguments for the `Tune` class.

    Attributes:
        objective: The location of the objective(s) to optimise for in the `Process`.
        parameters: The parameters to optimise over.
        num_samples: The number of samples to draw during the optimisation.
        mode: The mode of optimisation. For multi-objective optimisation, this should be a list
            containing a direction for each objective.
        max_concurrent: The maximum number of concurrent trials.
        algorithm: The algorithm to use for the optimisation.
    """

    objective: ObjectiveSpec | list[ObjectiveSpec]
    parameters: list[ParameterSpec] = Field(min_length=1)
    num_samples: PositiveInt
    mode: Direction | list[Direction] = "max"
    max_concurrent: PositiveInt | None = None
    algorithm: _t.Union[OptunaSpec] = Field(OptunaSpec(), discriminator="type")

    @model_validator(mode="after")
    def _validate_model(self: _t.Self) -> _t.Self:
        if isinstance(self.mode, list):
            if not isinstance(self.objective, list):
                raise ValueError(
                    "In multi-objective optimisation, both `mode` and `objective` must be lists."
                )
            if len(self.mode) != len(self.objective):
                raise ValueError("The length of `mode` must match the length of `objective`.")
        return self


class TuneSpec(PlugboardBaseModel):
    """Configuration for an optimisation job.

    Attributes:
        args: The arguments for the `Tune` job.
    """

    args: TuneArgsSpec



"""Provides spec classes related to `Connector`s."""

from collections.abc import Container
from enum import StrEnum
import re
import typing as _t

from pydantic import Field, field_validator

from ._common import PlugboardBaseModel


DEFAULT_CONNECTOR_CLS_PATH: str = "plugboard.connector.AsyncioConnector"


class ConnectorMode(StrEnum):
    """Defines the mode of a connector.

    Attributes:
        PIPELINE: one-in-one-out task queue.
        PUBSUB: one-to-many event distribution.
    """

    PIPELINE = "pipeline"
    PUBSUB = "pubsub"


class ConnectorSocket(PlugboardBaseModel):
    """`ConnectorSocket` defines a source or target connection point on a `Connector`.

    There are two typical types of connections in use: those between attributes of components;
    and those connecting components with events which they either emit or consume. When connecting
    two component attributes together, the `entity` is the name of the component, and the
    `descriptor` is the name of the attribute. When connecting components with events, the `entity`
    is the name of the event, and the `descriptor` is either "publishers" or "subscribers" as
    appropriate.

    Attributes:
        entity: The name of the entity.
        descriptor: The name of the descriptor on the entity.
    """

    _PATTERN: _t.ClassVar[re.Pattern] = re.compile(
        r"^([a-zA-Z_][a-zA-Z0-9_\-]*)\.([a-zA-Z_][a-zA-Z0-9_]*)$"
    )

    entity: str
    descriptor: str

    @classmethod
    def from_ref(cls, ref: str) -> _t.Self:
        """Creates a `ConnectorSocket` from a reference string."""
        match = cls._PATTERN.match(ref)
        if not match:
            raise ValueError(f"Reference must be of the form 'entity.descriptor', got {ref}")
        entity, descriptor = match.groups()
        return cls(entity=entity, descriptor=descriptor)

    @property
    def id(self) -> str:
        """Unique ID for `ConnectorSocket`."""
        return f"{self.entity}.{self.descriptor}"

    def __str__(self) -> str:
        return self.id

    def connects_to(self, entities: Container[str]) -> bool:
        """Returns `True` if the `ConnectorSocket` connects to any of the named entities."""
        return self.entity in entities


class ConnectorSpec(PlugboardBaseModel):
    """`ConnectorSpec` defines a connection between two entities.

    Attributes:
        source: The source endpoint.
        target: The target endpoint.
        mode: The mode of the connector.
    """

    source: ConnectorSocket
    target: ConnectorSocket
    mode: ConnectorMode = Field(default=ConnectorMode.PIPELINE, validate_default=True)

    @field_validator("source", "target", mode="before")
    @classmethod
    def _validate_source_target(cls, v: ConnectorSocket | dict | str) -> ConnectorSocket | dict:
        if isinstance(v, str):
            return ConnectorSocket.from_ref(v)
        return v

    @property
    def id(self) -> str:
        """Unique ID for `ConnectorSpec`."""
        return f"{self.source.id}..{self.target.id}"

    def __str__(self) -> str:
        return self.id


class ConnectorBuilderArgsDict(_t.TypedDict):
    """`TypedDict` of the [`Connector`][plugboard.connector.Connector] constructor arguments."""

    parameters: dict[str, _t.Any]


class ConnectorBuilderArgsSpec(PlugboardBaseModel, extra="allow"):
    """Specification of the [`Connector`][plugboard.connector.Connector] constructor arguments.

    Attributes:
        parameters: Parameters for the `Connector`.
    """

    parameters: dict[str, _t.Any] = {}


class ConnectorBuilderSpec(PlugboardBaseModel):
    """Specification of a `ConnectorBuilder`.

    Attributes:
        type: The type of the `ConnectorBuilder`.
        args: Optional; The arguments for the `ConnectorBuilder`.
    """

    type: str = DEFAULT_CONNECTOR_CLS_PATH
    args: ConnectorBuilderArgsSpec = ConnectorBuilderArgsSpec()
